//
//  Arduino Wavetable Sound Library
//
//  Copyright(C) 2010 JuJu
//    http://su-u.jp/juju/
//
//  History:
//    2010/06/24 Ver.0.1
//

#include <WTSound.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "wiring_private.h"
#include "pins_arduino.h"

#if F_CPU < 10000000
  #define WTSOUND_SPEED_SLOW
#endif

// ROM wave tables
#define WTSOUND_WAVE_TABLE_ROMS 5
const prog_int8_t waveTableROM[WTSOUND_WAVE_TABLE_ROMS][WTSOUND_WAVE_TABLE_SIZE] PROGMEM = 
{
  {
      0,  3,  6,  8, 11, 12, 14, 15, 15, 15, 14, 12, 11,  8,  6,  3,
      0, -3, -6, -8,-11,-12,-14,-15,-15,-15,-14,-12,-11, -8, -6, -3, // Sine wave
  },
  {
      0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
      0,-15,-14,-13,-12,-11,-10, -9, -8, -7, -6, -5, -4, -3, -2, -1, // Sawtooth wave
  },
  {
      0,  2,  4,  6,  8, 10, 12, 14, 16, 14, 12, 10,  8,  6,  4,  2,
      0, -2, -4, -6, -8,-10,-12,-14,-16,-14,-12,-10, -8, -6, -4, -2, // Triangle wave
  },
  {
     15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
    -15,-15,-15,-15,-15,-15,-15,-15,-15,-15,-15,-15,-15,-15,-15,-15, // Square wave
  },
  {
     -7, -1,  0, 10,  5, 15,  7,  8,  9,  0,  4, 13, 10,  1,  1, -1,
      1,  5, -8,  3, -6, -2,-10, -2, -4,-12,-15, -1, -2, -9, -4, -3, // Noise
  },
};

// Note width (21 - 108)
uint16_t noteWidth[] =
{
      0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
      0,    0,    0,    0,    0,  225,  239,  253,  268,  284,  301,  319,  338,  358,  379,  401,
    425,  451,  477,  506,  536,  568,  601,  637,  675,  715,  758,  803,  851,  901,  955, 1011,
   1072, 1135, 1203, 1274, 1350, 1430, 1515, 1606, 1701, 1802, 1909, 2023, 2143, 2271, 2406, 2549,
   2700, 2861, 3031, 3211, 3402, 3604, 3819, 4046, 4286, 4541, 4811, 5098, 5401, 5722, 6062, 6422,
   6804, 7209, 7638, 8092, 8573, 9083, 9623,10195,10801,11444,12124,12845,13609,14418,15275,16184,
  17146,18165,19246,20390,21602,22887,24248,25690,27217,28836,30551,32367,34292,    0,    0,    0,
      0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
};

WTSoundClass::WTSoundClass()
{
  for (uint8_t i = 0; i < WTSOUND_CHANNELS; i++)
  {
    _count[i] = 0;
    _countadd[i] = 0;
    _note[i] = 0;
    _velocity[i] = 127;
    memcpy_P(_waveTable[i], waveTableROM[1], WTSOUND_WAVE_TABLE_SIZE);
  }
  _velocityPCM = 127;
  _tempo = 0;
  _funcInterval = NULL;
  _countIntervalT = 0;
}

void WTSoundClass::begin(uint8_t pin)
{
  begin(pin, NULL, 0);
}

void WTSoundClass::begin(uint8_t pin, void* bufferPCM, uint16_t bufferPCMSize)
{
  digitalWrite(pin, LOW);
  pinMode(pin, OUTPUT);
  // Init wave buffer
  _bufferPCM = (int8_t *)bufferPCM;
  _bufferPCMSize = bufferPCMSize;
  _bufferPCMHead = 0;
  _bufferPCMTail = 0;
  // Init timer
  #ifdef TIMSK2
    TIMSK2 = 0x00;
    TCCR2A = _BV(WGM21) | _BV(WGM20);
    TCCR2B = _BV(WGM22) | _BV(CS20);
    #ifndef WTSOUND_SPEED_SLOW
      OCR2A  = (uint8_t)(F_CPU / 80000);
      OCR2B  = (uint8_t)(F_CPU / 80000 / 2);
    #else
      OCR2A  = (uint8_t)(F_CPU / 40000);
      OCR2B  = (uint8_t)(F_CPU / 40000 / 2);
    #endif
    if (digitalPinToTimer(pin) == TIMER2A)
      TCCR2A |= _BV(COM2A1);
    if (digitalPinToTimer(pin) == TIMER2B)
      TCCR2A |= _BV(COM2B1);
    TIMSK2 = _BV(TOIE2);
  #endif
}

void WTSoundClass::setInterval(uint16_t tempo)
{
  if (_tempo != tempo)
  {
    _tempo = tempo;
    if (tempo > 0)
      #ifndef WTSOUND_SPEED_SLOW
        _countIntervalT = 16000 / tempo;
      #else
        _countIntervalT = 8000 / tempo;
      #endif
    else
      _countIntervalT = 0;
    _countInterval = _countIntervalT;
  }
}

void WTSoundClass::setInterval(uint16_t tempo, void (* func)())
{
  _funcInterval = func;
  setInterval(tempo);
}

void WTSoundClass::setVelocity(uint8_t ch, int8_t velocity)
{
  if (ch < WTSOUND_CHANNELS && velocity >= 0)
      _velocity[ch] = velocity;
}

void WTSoundClass::setProgram(uint8_t ch, uint8_t prog)
{
  if (ch < WTSOUND_CHANNELS && prog < WTSOUND_WAVE_TABLE_ROMS)
    memcpy_P(_waveTable[ch], waveTableROM[prog], WTSOUND_WAVE_TABLE_SIZE);
}

void WTSoundClass::setWaveTable(uint8_t ch, int8_t *wave)
{
  if (ch < WTSOUND_CHANNELS)
    memcpy(_waveTable[ch], wave, WTSOUND_WAVE_TABLE_SIZE);
}

void WTSoundClass::setVelocityPCM(int8_t velocity)
{
  if (velocity >= 0)
      _velocityPCM = velocity;
}

void WTSoundClass::play(uint8_t ch, uint8_t note)
{
  if (ch < WTSOUND_CHANNELS && note <= 127)
  {
    if (_note[ch] != note)
    {
      _note[ch] = note;
      uint16_t width = noteWidth[note];
      #ifdef WTSOUND_SPEED_SLOW
        width *= 2;
      #endif
      _countadd[ch] = width;
      if (width == 0)
        _count[ch] = 0;
    }
  }
}

void WTSoundClass::play(uint8_t ch, uint8_t note, int8_t velocity)
{
  setVelocity(ch, velocity);
  play(ch, note);
}

void WTSoundClass::playPCM(int8_t *wave, uint16_t size)
{
  if (_bufferPCMSize > 0)
  {
    uint16_t head = _bufferPCMHead;
    for ( ; size > 0; size--)
    {
      if (++head >= _bufferPCMSize)
        head = 0;
      while (head == _bufferPCMTail)
        ;
      _bufferPCM[head] = *wave++;
    }
    _bufferPCMHead = head;
  }
}

void WTSoundClass::_interval()
{
  int16_t level = 0;
  // Play notes
  for (uint8_t i = 0; i < WTSOUND_CHANNELS; i++)
  {
    if (_note[i] > 0)
    {
      _count[i] +=_countadd[i];
      level += (int16_t)_waveTable[i][_count[i] >> 11] * (int8_t)_velocity[i];
    }
  }
  level *= 8;
  // Play PCM
  if (_bufferPCMHead != _bufferPCMTail)
  {
    #ifndef WTSOUND_SPEED_SLOW
    static uint8_t wait_pcm = 1;
    if (--wait_pcm == 0)
    {
      wait_pcm = 2;
    #else
    {
    #endif
      if (++_bufferPCMTail >= _bufferPCMSize)
        _bufferPCMTail = 0;
      level = (int16_t)_bufferPCM[_bufferPCMTail] * (int8_t)_velocityPCM;
    }
  }
  // PWM set
  #ifdef TIMSK2
    if (-16384 <= level && level <= 16383)
    {
      #ifndef WTSOUND_SPEED_SLOW
        OCR2B = (uint8_t)(F_CPU / 80000 / 2)
              + (((level >> 7) * (uint8_t)(F_CPU / 80000)) >> 8);
      #else
        OCR2B = (uint8_t)(F_CPU / 40000 / 2)
              + (((level >> 7) * (uint8_t)(F_CPU / 40000)) >> 8);
      #endif
    }
  #endif
  // Tempo
  if (_countIntervalT > 0 && --_countInterval == 0)
  {
    _countInterval = _countIntervalT;
    if (_funcInterval != NULL)
      (*_funcInterval)();
  }
}

WTSoundClass WTSound;

#ifdef TIMSK2
  ISR(TIMER2_OVF_vect)
  {
    static uint8_t wait = 1;
    if (--wait == 0)
    {
      wait = 5;
      TIMSK2 |= _BV(OCIE2A);
    }
  }
  ISR(TIMER2_COMPA_vect)
  {
    static volatile bool exec = false;
    TIMSK2 &= ~_BV(OCIE2A);
    if (!exec)
    {
      exec = true;
      sei();
      WTSound._interval();
      exec = false;
    }
  }
#endif
